import streamlit as st
from io import BytesIO
from PIL import Image
from functions.chatbot_api.agent import run_agent, run_detection
from visualization import draw_graph
from functions.chatbot_api.chains.assessment_chain import clear_patient_context_cache
from functions.chatbot_api.chains.behavioral_chain import clear_comparison_cache
from functions.chatbot_api.chains.diagnosis_chain import clear_patient_diagnosis_cache
from functions.chatbot_api.chains.prognosis_chain import clear_prognosis_cache
from functions.chatbot_api.chains.treatment_chain import clear_treatment_plan_cache

from langchain_core.messages import ToolMessage
import re
import json

# Constants
SAMPLE_REPORT = "FINDINGS : Normal heart size and mediastinal contours. There is interval improvement in the right lower lobe airspace disease. There is XXXX mild air space opacity in that distribution. No XXXX airspace disease. No pneumothorax or pleural effusion. IMPRESSION : Improved right lower lobe consolidation with mild residual or recurrent pneumonia. Recommend radiographic follow up after appropriate therapy to confirm resolution. If there is no improvement consider XXXX for further evaluation."
PAGE_TITLE = "Medical Copilot"
PAGE_ICON = "favicon.ico"
AREAS = [
    "Detection",
    "Assessment",
    "Diagnosis",
    "Prognosis",
    "Treatment Plan",
    # "Behavioral Insights", Commented out for supermicro demo because it requires Neo4J
]


def initialize_session_state():
    for key, default_value in {
        "messages": [],
        "monai_summary": None,
        "patient_ssn": "",
        "report": "",
        "monai_detection": None,
        "detection_performed": False,
        "prognosis": None,
    }.items():
        if key not in st.session_state:
            st.session_state[key] = default_value


def clear_chat_history():
    st.session_state.messages = []
    st.session_state.monai_summary = None
    st.session_state.monai_detection = None
    st.session_state.detection_performed = False
    st.session_state.prognosis = None
    clear_patient_context_cache()
    clear_comparison_cache()
    clear_patient_diagnosis_cache()
    clear_prognosis_cache()
    clear_treatment_plan_cache()


def display_sidebar():
    with st.sidebar:
        # Hopsworks Logo
        st.image(image="images/hopsworks_logo.png")

        # Main content
        st.markdown(
            "<h1 style='text-align: right; margin-top: -52px; margin-right: 23px'>Medical Copilot</h1>",
            unsafe_allow_html=True,
        )

        st.session_state.patient_ssn = st.text_input(
            "🪪 Patient SSN", value=st.session_state.patient_ssn
        )

        area = st.selectbox(
            "Which areas are you interested in asking questions about", AREAS, index=0
        )

        if area == "Behavioral Insights":
            display_behavioral_insights_examples()
        elif area == "Detection":
            handle_detection_area()
        elif area in ["Treatment Plan", "Assessment", "Diagnosis", "Prognosis"]:
            display_example_questions(area)

    return area


def display_behavioral_insights_examples():
    st.header("📖 Example Questions:")
    examples = [
        "Can you show me the journey of a patient who has been diagnosed with pneumonia?",
        "I'd like to see the care pathway for a patient who has been readmitted to the hospital. Can you show me that?",
        "What are the most common reasons patients get readmitted to the hospital within 30 days?",
        "Can you show me patients with condition code 444814009?",
        "Compare similar patients with and without a pneumonia",
        "Compare patients with the lung opacity where the clinical notes are anomalous",
    ]
    for example in examples:
        st.markdown(f"- {example}")


def handle_detection_area():
    st.header("🩻 Detection")

    uploaded_file = st.file_uploader("🖼️ Upload X-Ray Image:")
    if uploaded_file:
        image_object = Image.open(BytesIO(uploaded_file.getvalue())).resize((512, 512))
        st.image(image_object)

    st.session_state.report = st.text_area("📋 Technician Report", SAMPLE_REPORT)

    if st.button("Detect", type="primary"):
        process_detection(image_object if "image_object" in locals() else None)

    st.markdown(f"- Explain the key points of lung opacity.")


def process_detection(image_object):
    clear_chat_history()
    if image_object:
        with st.spinner("🔎 Running detection..."):
            monai_detection = run_detection(image_object, st.session_state.report)
            print(monai_detection)
            st.session_state.monai_detection = monai_detection
            # The monai model can predict no_finding and support_devices which
            # which are then problematic for the next iteractions as the LLM
            # might think they are a condition
            # Here we remove them from the state
            del st.session_state.monai_detection["no_finding"]
            del st.session_state.monai_detection["support_devices"]

            # Summarize the detection (including potentially no findings)
            process_monai_detection(monai_detection)

            st.session_state.detection_performed = True
    else:
        st.error("💾 Please upload an X-Ray image before detection.")


def process_monai_detection(monai_detection):
    if monai_detection is not None:
        comprehensive_analysis_prompt = """
        Provide a comprehensive, data-driven analysis of the findings based on the MONAI model predictions and the radiologist's report.
        """
        with st.spinner("⚙️ Generating comprehensive analysis..."):
            response = run_agent(
                user_query=comprehensive_analysis_prompt,
                monai_detection=json.dumps(st.session_state.monai_detection),
                patient_ssn=st.session_state.patient_ssn,
                report_text=st.session_state.report,
            )

        if isinstance(response["output"], ToolMessage):
            st.session_state.monai_summary = response["output"].content
        else:
            st.session_state.monai_summary = str(response["output"])

        st.success(
            "✅ Detection and analysis complete. You can now ask questions about the results."
        )
    else:
        st.error("❌ Detection failed. Please try again.")


def display_example_questions(area):
    st.header("📖 Example Questions:")
    examples = {
        "Treatment Plan": [
            "Does the patient have a treatment plan?",
            "If no existing plan, create a comprehensive treatment plan, based on all available information about the patient.",
            "If there is an existing treatment plan, have there been any changes since their last primary care visit? If yes, describe these changes and create an updated treatment plan. If no changes, confirm the current plan is still appropriate.",
            "Outline the treatment plan for the following time periods: a. Next 3 months; b. Next 6 months; c. Next 9 months; d. Next 12 months",
            "What is the next immediate step in the patient's care?",
            "What are some example treatment plans for patients where a particular condition is detected?",
            "Compare the patient's prior treatment plan with expected treatment plan",
        ],
        "Assessment": [
            "Tell me everything I know about the patient",
            "What are the risk factors for this patient?",
            "Is there sufficient data to create a screening and workup plan?",
            "What additional information would be most valuable to improve the screening and workup process?",
            "Using only the currently available information, what would be the best possible screening and workup plan?",
            "Are there any new observations since the last visit?",
            "What other questions should I ask the patient in order to make a more thorough assessment?",
            "Which tests should I order to check for a condition?",
        ],
        "Diagnosis": [
            "Do I have any observations for the patient?",
            "Do I have enough information to diagnose any conditions? If so, what is the diagnosis? If not, what information do I need?",
            "What other questions should I ask the patient in order to make a more accurate diagnosis?",
        ],
        "Prognosis": [
            "Do I have enough information to make a reliable prognosis? If so, what is it? If not, what information do I need?",
            "Create a prognosis for my patient",
            "Compare the prognosis of a patient with the prognosis of similar patients with a particular attribute",
            "What other questions should I ask the patient in order to make a more accurate prognosis?",
        ],
    }
    for example in examples.get(area, []):
        st.markdown(f"- {example}")


def display_chat_history():
    for message in st.session_state.messages:
        with st.chat_message(message["role"]):
            if message["role"] == "user":
                st.markdown(
                    f'<p style="font-size: 20px;">{message["content"]}</p>',
                    unsafe_allow_html=True,
                )
            elif "dataframe" in message:
                st.dataframe(message["dataframe"], hide_index=True)
            elif "graph" in message:
                try:
                    st.write(draw_graph(message["graph"]))
                except Exception as e:
                    st.error(f"Error displaying graph: {str(e)}")
                    st.json(message["graph"])
            elif "content" in message:
                st.markdown(message["content"])
            else:
                st.warning("Message content not found.")


def handle_user_query():
    user_query = st.chat_input("What do you want to know?")
    if user_query:
        st.chat_message("user").markdown(
            f'<p style="font-size: 20px;">{user_query}</p>', unsafe_allow_html=True
        )
        st.session_state.messages.append({"role": "user", "content": user_query})

        if not st.session_state.patient_ssn:
            display_ssn_warning()
            return

        if not st.session_state.detection_performed:
            display_detection_warning()
            return

        with st.spinner("⚙️ Generating response..."):
            response = run_agent(
                user_query=user_query,
                monai_detection=st.session_state.monai_detection,
                patient_ssn=st.session_state.patient_ssn,
                report_text=st.session_state.report,
            )

        display_response(response)


def display_ssn_warning():
    warning_message = (
        "⚠️ Please enter a Patient SSN in the sidebar before asking questions."
    )
    st.warning(warning_message)
    st.session_state.messages.append({"role": "assistant", "content": warning_message})


def display_detection_warning():
    warning_message = "⚠️ Please perform detection first before asking questions. Upload an X-Ray image and click 'Detect' in the sidebar."
    st.warning(warning_message)
    st.session_state.messages.append({"role": "assistant", "content": warning_message})


def display_response(response):
    output = response["output"]

    def adjust_markdown_headers(content):
        content = re.sub(r"^# ", "## ", content, flags=re.MULTILINE)
        content = re.sub(r"^## ", "### ", content, flags=re.MULTILINE)
        return content

    if isinstance(output, str):
        adjusted_content = adjust_markdown_headers(output)
        st.markdown(adjusted_content)
        st.session_state.messages.append(
            {"role": "assistant", "content": adjusted_content}
        )

    elif hasattr(output, "name"):
        if output.name == "BehavioralQueryAssistant":
            if (
                hasattr(output, "response_metadata")
                and "visualization" in output.response_metadata
            ):
                visualization = output.response_metadata["visualization"]
                if visualization == "table":
                    st.dataframe(output.artifact, hide_index=True)
                    st.session_state.messages.append(
                        {"role": "assistant", "dataframe": output.artifact}
                    )
                elif visualization == "graph":
                    try:
                        st.write(draw_graph(output.artifact))
                        st.session_state.messages.append(
                            {"role": "assistant", "graph": output.artifact}
                        )
                    except Exception as e:
                        st.error(f"Error displaying graph: {str(e)}")
                        st.json(output.artifact)
                        st.session_state.messages.append(
                            {
                                "role": "assistant",
                                "content": "Error displaying graph. Raw data shown.",
                            }
                        )
            else:
                # If there's no visualization metadata, display the content as text
                adjusted_content = adjust_markdown_headers(output.content)
                st.markdown(adjusted_content)
                st.session_state.messages.append(
                    {"role": "assistant", "content": adjusted_content}
                )
        elif output.name == "PrognosisAssistant":
            adjusted_content = adjust_markdown_headers(output.content)
            st.markdown(adjusted_content)
            st.session_state.messages.append(
                {"role": "assistant", "content": adjusted_content}
            )
            st.session_state.prognosis = output.content
        else:
            adjusted_content = adjust_markdown_headers(output.content)
            st.markdown(adjusted_content)
            st.session_state.messages.append(
                {"role": "assistant", "content": adjusted_content}
            )

    elif isinstance(output, ToolMessage):
        adjusted_content = adjust_markdown_headers(output.content)
        st.markdown(adjusted_content)
        st.session_state.messages.append(
            {"role": "assistant", "content": adjusted_content}
        )

    else:
        adjusted_content = adjust_markdown_headers(str(output))
        st.markdown(adjusted_content)
        st.session_state.messages.append(
            {"role": "assistant", "content": adjusted_content}
        )


def main():
    st.set_page_config(page_title=PAGE_TITLE, page_icon=PAGE_ICON, layout="wide")

    initialize_session_state()
    area = display_sidebar()

    if st.session_state.monai_summary:
        st.write("🩻 Detection and Analysis Summary:")
        st.write(st.session_state.monai_summary)

    display_chat_history()
    handle_user_query()


if __name__ == "__main__":
    main()
