import base64
import hopsworks

from typing import Annotated
from typing_extensions import TypedDict

from langchain_core.tools import tool
from langchain_core.messages import ToolMessage
from langchain_core.prompts.chat import ChatPromptTemplate, MessagesPlaceholder
from langgraph.graph import StateGraph, START, END
from langgraph.graph.message import add_messages
from langchain_core.runnables.config import RunnableConfig

import settings

from functions.chatbot_api.llm_config import get_llm
from functions.chatbot_api.chains.detection_chain import DetectionChain
from functions.chatbot_api.chains.prognosis_chain import PrognosisChain
from functions.chatbot_api.chains.treatment_chain import TreatmentChain
from functions.chatbot_api.chains.behavioral_chain import BehavioralChain
from functions.chatbot_api.chains.assessment_chain import assessment_assistant
from functions.chatbot_api.chains.diagnosis_chain import diagnosis_assistant


# Custom Memory Component
# class CustomConversationBufferMemory(ConversationBufferMemory):
#     def save_context(self, inputs: dict, outputs: dict) -> None:
#         input_str, output = self._get_input_output(inputs, outputs)
#         self.chat_memory.add_user_message(input_str)
#
#         if isinstance(output, ToolMessage):
#             self.chat_memory.add_ai_message(output.content)
#         elif isinstance(output, str):
#             self.chat_memory.add_ai_message(output)
#         else:
#             self.chat_memory.add_ai_message(str(output))
#
#     def _get_input_output(self, inputs: dict, outputs: dict) -> tuple:
#         prompt_input_key = self.input_key if self.input_key else "input"
#         output_key = self.output_key if self.output_key else "output"
#         return inputs[prompt_input_key], outputs[output_key]


# Hopsworks and Model Setup
project = hopsworks.login()
fs = project.get_feature_store()

# Retrieve Reports Feature Group
report_fg = fs.get_feature_group(
    name="report",
    version=5,
)

# Retrieve Feature Views
outcomes_fv = fs.get_feature_view(
    name="outcomes",
    version=1,
)
report_fv = fs.get_feature_view(
    name="reports_detection",
    version=1,
)
patient_data_fv = fs.get_feature_view(
    "patient_data",
    version=2,
)
historical_observations_fv = fs.get_feature_view(
    "historical_observations",
    version=2,
)
latest_observations_fv = fs.get_feature_view(
    "latest_observations",
    version=2,
)
prognoses_fv = fs.get_feature_view(
    "prognoses",
    version=1,
)
treatment_plans_fv = fs.get_feature_view(
    name="treatment_plans",
    version=1,
)

ms = project.get_model_serving()
deployment = ms.get_deployment(settings.DEPLOYMENT_NAME)

# Initialize DetectionChain
detection_chain = DetectionChain(
    report_fg,
    outcomes_fv,
    report_fv,
    patient_data_fv,
    historical_observations_fv,
    latest_observations_fv,
)


@tool
def detection_assistant_wrapper(config: RunnableConfig) -> str:
    """
    Use this tool for queries on medical imaging and specific test results:
    - Analyze image findings
    - Interpret specific test results
    - Classify images
    - Provide context-based summaries
    Not for general observations or population analysis.
    """
    user_query = config.get("configurable", {}).get("user_query")
    monai_detection = config.get("configurable", {}).get("monai_detection")
    report_text = config.get("configurable", {}).get("report_text")

    return detection_chain(user_query, monai_detection, report_text)


prognosis_chain = PrognosisChain(
    outcomes_fv,
    report_fv,
    prognoses_fv,
    report_fg,
)


@tool
def prognosis_assistant_wrapper(config: RunnableConfig) -> ToolMessage:
    """
    For patient prognosis queries:
    - Evaluate prognosis information
    - Provide prognosis based on data
    - Compare with similar cases
    - Suggest prognosis-specific questions
    - Identify factors affecting outlook
    - Recommend additional prognosis info
    Always use when 'prognosis' is mentioned.
    Not for general assessment or treatment.
    """
    chat_history = ""
    user_query = config.get("configurable", {}).get("user_query")
    patient_ssn = config.get("configurable", {}).get("patient_ssn")
    monai_detection = config.get("configurable", {}).get("monai_detection")

    return prognosis_chain(
        user_query=user_query,
        patient_ssn=patient_ssn,
        chat_history=chat_history,
        monai_detection=monai_detection,
    )


# Initialize TreatmentChain
treatment_chain = TreatmentChain(
    outcomes_fv,
    treatment_plans_fv,
)


@tool
def treatment_assistant_wrapper(config: RunnableConfig) -> str:
    """
    For individual patient treatment/care:
    - Check existing plans
    - Create/update treatment plans
    - Assess condition changes affecting treatment
    - Outline plans for various periods
    - Determine next care steps
    - What are some example treatment plans for patients where a particular condition is detected?
    - Compare current/prior plans
    Use for 'treatment', 'plan' queries.
    Not for population-level analysis.
    """
    chat_history = ""
    user_query = config.get("configurable", {}).get("user_query")
    patient_ssn = config.get("configurable", {}).get("patient_ssn")
    monai_detection = config.get("configurable", {}).get("monai_detection")

    return treatment_chain(
        user_query=user_query,
        monai_detection=monai_detection,
        patient_ssn=patient_ssn,
        chat_history=chat_history,
    )


behavioral_chain = BehavioralChain(
    outcomes_fv,
    report_fv,
    patient_data_fv,
    historical_observations_fv,
    report_fg,
    latest_observations_fv,
)


def query_behavioral_chain(user_query: str) -> ToolMessage:
    chat_history = ""
    return behavioral_chain(user_query, chat_history)


@tool
def assessment_assistant_wrapper(config: RunnableConfig) -> str:
    """
    For general patient assessment:
    - Summarize patient information
    - Identify risk factors
    - Create/evaluate screening plans
    - Recommend additional tests/info
    - Analyze general condition changes
    - Suggest assessment questions
    Not for specific diagnosis, observations, prognosis, or treatment.
    """
    user_query = config.get("configurable", {}).get("user_query")
    patient_ssn = config.get("configurable", {}).get("patient_ssn")
    monai_detection = config.get("configurable", {}).get("monai_detection")
    return assessment_assistant(
        user_query=user_query,
        monai_detection=monai_detection,
        patient_ssn=patient_ssn,
        patient_data_fv=patient_data_fv,
        historical_observations_fv=historical_observations_fv,
        latest_observations_fv=latest_observations_fv,
    )


@tool
def diagnosis_assistant_wrapper(config: RunnableConfig) -> str:
    """
    For patient diagnosis and observations:
    - List/evaluate current observations
    - Assess diagnosis information sufficiency
    - Suggest info needed for diagnosis
    - Provide potential diagnoses
    - Recommend diagnosis-specific questions
    Always use for "Do I have any observations for the patient?"
    Not for general assessment, prognosis, or treatment.
    """
    user_query = config.get("configurable", {}).get("user_query")
    patient_ssn = config.get("configurable", {}).get("patient_ssn")
    monai_detection = config.get("configurable", {}).get("monai_detection")
    return diagnosis_assistant(
        user_query=user_query,
        patient_id=patient_ssn,
        historical_observations_fv=historical_observations_fv,
        latest_observations_fv=latest_observations_fv,
        monai_detection=monai_detection,
    )


# Set up the tools
tools = [
    detection_assistant_wrapper,
    prognosis_assistant_wrapper,
    treatment_assistant_wrapper,
    assessment_assistant_wrapper,
    diagnosis_assistant_wrapper,
]
tools_by_name = {tool.name: tool for tool in tools}

# LLM Setup
chat_model = get_llm(
    llm_type=settings.LLM_TYPE,
    temperature=settings.AGENT_LLM_TEMPERATURE,
).bind_tools(tools)

# Prompt Setup
system_message = """
You are an AI medical assistant primarily designed to help with medical queries and patient care. Your main goal is to provide accurate and relevant information about patients and their medical conditions. Follow these guidelines:

1. For all queries:
   a. If the user provides a brief personal greeting, acknowledge it politely but concisely, then immediately focus on medical topics.
   b. Always emphasize your role as a medical assistant and your readiness to help with health-related questions or concerns.

2. For medical queries:
   a. Use the AssessmentAssistant tool for questions related to a specific, individual patient's assessment, including risk factors, screening plans, or specific patient information that doesn't fall under other categories. For the observations SINCE THE LAST VISIT invoke the AssessmentAssistant.
   b. Use the TreatmentAssistant tool only for specific questions about treatments or treatment plans for an individual patient. ALWAYS invoke the TreatmentAssistant if the phrase "treatment plan" is present in the user query.
   c. Use the DetectionAssistant tool for questions related to medical image analysis, detection predictions, and interpretation of X-ray reports for individual patients.
   d. Use the DiagnosisAssistant tool for questions about a specific patient's observations, determining if there's enough information for a diagnosis, or suggesting additional questions for diagnosis.
   e. Use the PrognosisAssistant tool for questions about a specific patient's prognosis, determining if there's enough information for a prognosis, or suggesting additional questions for a more accurate prognosis. IMPORTANT: If the word "prognosis" is present in the user query, ALWAYS use the PrognosisAssistant tool.
   f. Use the BehavioralQueryAssistant tool for queries about aggregated patient data, population-level insights, common reasons patients get readmitted, general healthcare trends, patient journeys, care pathways, or readmission patterns that are not specific to an individual patient. This includes comparing groups of patients (e.g., with and without specific conditions, including ANOMALOUS notes) and analyzing reasons for readmissions.

   When using any medical tool, you MUST provide all required inputs.

   If a query involves multiple patients, general trends, or patterns (especially related to care pathways or readmissions), use the BehavioralQueryAssistant.

3. For non-medical queries or general conversation:
   a. Briefly acknowledge the query if it's a greeting or introduction.
   b. Immediately redirect the conversation to your purpose as a medical assistant.
   c. Encourage the user to ask about any health concerns or medical questions they might have.

Remember, your primary function is to assist with medical queries. While you can be polite and acknowledge greetings, your responses should always be focused on providing medical assistance and encouraging health-related discussions.
"""
prompt = ChatPromptTemplate.from_messages(
    [
        ("system", system_message),
    ]
)


class State(TypedDict):
    messages: Annotated[list, add_messages]


def tool_node(state: dict):
    result = []
    for tool_call in state["messages"][-1].tool_calls:
        tool = tools_by_name[tool_call["name"]]
        result.append(tool.invoke(tool_call["args"]))
    return {"messages": result}


def chatbot(state: State):
    print(state["messages"])
    return {"messages": [chat_model.invoke(state["messages"])]}


def route_tools(
    state: State,
):
    """
    Use in the conditional_edge to route to the ToolNode if the last message
    has tool calls. Otherwise, route to the end.
    """
    if isinstance(state, list):
        ai_message = state[-1]
    elif messages := state.get("messages", []):
        ai_message = messages[-1]
    else:
        raise ValueError(f"No messages found in input state to tool_edge: {state}")
    if hasattr(ai_message, "tool_calls") and len(ai_message.tool_calls) > 0:
        return "tools"
    return END


graph_builder = StateGraph(State)

graph_builder.add_node("tools", tool_node)
graph_builder.add_node("chatbot", chatbot)
graph_builder.add_conditional_edges(
    "chatbot",
    route_tools,
    # The following dictionary lets you tell the graph to interpret the condition's outputs as a specific node
    # It defaults to the identity function, but if you
    # want to use a node named something else apart from "tools",
    # You can update the value of the dictionary to something else
    # e.g., "tools": "my_tools"
    {"tools": "tools", END: END},
)
graph_builder.add_edge(START, "chatbot")
graph_builder.add_edge("chatbot", END)
graph = graph_builder.compile()


def run_agent(
    user_query: str,
    monai_detection: str,
    patient_ssn: str,
    report_text: str,
):
    print(f"📋 Patient SSN: {patient_ssn}")

    answer = graph.invoke(
        {"messages": [("system", system_message), ("user", user_query)]},
        {
            "configurable": {
                "patient_ssn": patient_ssn,
                "monai_detection": monai_detection,
                "report_text": report_text,
                "user_query": user_query,
            }
        },
    )

    return {"output": answer["messages"][-1].content}


def run_detection(image, report):
    return deployment.predict(
        inputs=[
            {
                "report": report,
                "image": base64.b64encode(image.tobytes()).decode("utf-8"),
            }
        ]
    )['predictions']
