import logging
from typing import Dict, Any, Optional
from langchain.prompts import PromptTemplate
from langchain_core.messages import ToolMessage
import settings
from functions.chatbot_api.llm_config import get_llm
import json
import pandas as pd
import numpy as np

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class PatientContextCache:
    def __init__(self):
        self.cache = {}

    def get(self, patient_ssn: str) -> Optional[Dict[str, Any]]:
        if patient_ssn in self.cache:
            logger.info(f"✅ Cache hit for patient SSN: {patient_ssn}")
            return self.cache[patient_ssn]
        logger.info(f"🔎 Cache miss for patient SSN: {patient_ssn}")
        return None

    def set(self, patient_ssn: str, context: Dict[str, Any]) -> None:
        logger.info(f"💾 Caching context for patient SSN: {patient_ssn}")
        self.cache[patient_ssn] = context

    def clear(self) -> None:
        logger.info("🧹 Clearing patient context cache")
        self.cache.clear()


patient_context_cache = PatientContextCache()


def get_patient_context(
    monai_detection: str,
    patient_ssn: int,
    patient_data_fv: Any,
    historical_observations_fv: Any,
    latest_observations_fv: Any,
) -> Dict[str, Any]:
    if patient_ssn == "":
        return f"❌ Wrong patient ssn: {patient_ssn}"

    # Retrieve cached patient context if present
    cached_context = patient_context_cache.get(str(patient_ssn))
    if cached_context:
        return cached_context

    logger.info(
        f"🪝 Retrieving patient context for SSN from the Feature Store: {patient_ssn}"
    )

    # Retrieve patient data
    patient_data = patient_data_fv.get_batch_data()

    patient_ssn = int(patient_ssn)

    patient_data_filtered = patient_data[patient_data.patient_id == patient_ssn]

    if patient_data_filtered.empty:
        return {
            "error": f"No patient data found for SSN: {patient_ssn}. Please check the patient ID and try again."
        }

    # Sort patient data by report_date and get the last 3 reports
    patient_data_last3 = patient_data_filtered.sort_values(
        "report_date", ascending=False
    ).head(3)

    # Retrieve historical observations
    historical_observations = historical_observations_fv.get_batch_data()
    patient_observations = historical_observations[
        historical_observations.patient_id == patient_ssn
    ]
    patient_observations_latest5 = patient_observations.sort_values(
        "observation_date", ascending=False
    ).head(5)

    # Get latest new observation
    try:
        latest_observation = latest_observations_fv.get_feature_vector(
            {"patient_id": patient_ssn}
        )
    except Exception as e:
        logger.error(f"Error retrieving latest observation: {str(e)}")
        latest_observation = None

    # Prepare patient context
    context = prepare_patient_context(
        patient_data_last3,
        patient_observations_latest5,
        latest_observation,
        monai_detection,
    )

    # Cache the patient context
    patient_context_cache.set(str(patient_ssn), context)

    return context


def prepare_patient_context(
    patient_data_last3,
    patient_observations_latest5,
    latest_observation,
    monai_detection,
):
    def format_observations(obs_string: str) -> list:
        return [o.strip() for o in obs_string.split("|")]

    def format_date(date):
        if isinstance(date, pd.Timestamp):
            return date.isoformat()
        return str(date)

    def simplify(obj):
        if isinstance(obj, (int, float, bool, str)):
            return obj
        return str(obj)

    print("================================================")
    print(latest_observation)
    print("================================================")

    return {
        "patient_id": int(patient_data_last3["patient_id"].iloc[0]),
        "monai_detection": monai_detection,
        "reports": [
            {
                "report_date": format_date(row["report_date"]),
                "findings": {
                    k: simplify(v)
                    for k, v in row.items()
                    if k not in ["patient_id", "report", "report_date", "report_text"]
                },
                "report_text": str(row.get("report_text", "No report available")),
            }
            for _, row in patient_data_last3.iterrows()
        ],
        "historical_observations": [
            {
                "date": format_date(row["observation_date"]),
                "observations": format_observations(str(row["observations"])),
            }
            for _, row in patient_observations_latest5.iterrows()
        ],
        "latest_observation_since_last_visit": {
            "date": format_date(latest_observation[1])
            if latest_observation
            else "No date available",
            "observations": format_observations(str(latest_observation[2]))
            if latest_observation
            else ["No new observations"],
        }
        if latest_observation
        else {"date": "No date available", "observations": ["No new observations"]},
    }


def assessment_assistant(
    user_query: str,
    monai_detection: str,
    patient_ssn: int,
    patient_data_fv: Any,
    historical_observations_fv: Any,
    latest_observations_fv: Any,
) -> str:
    """Function to answer user queries based on patient context."""
    logger.info(f"👨🏻‍⚕️ Assessment query for patient SSN: {patient_ssn}")

    patient_context = get_patient_context(
        monai_detection,
        patient_ssn,
        patient_data_fv,
        historical_observations_fv,
        latest_observations_fv,
    )

    if "error" in patient_context:
        return patient_context["error"]

    assessment_prompt_template = """
You are an advanced AI trained to support doctors in assessing patient conditions based on medical reports and historical data.
TASK: Answer the user's query based on the provided patient context.

<INSTRUCTIONS>
- Utilize ALL provided patient context to answer queries comprehensively, including:
  * Last 3 reports
  * MONAI summary
  * Historical observations
  * Observations since the last visit
  * Any other relevant information in the context
- For comprehensive assessments, provide a detailed overview of the patient's condition, noting changes across reports
- Analyze trends and changes in the patient's condition over time
- For risk factor analysis, consider both current findings and historical data
- When discussing screening and workup plans, factor in all available information and note any limitations
- Highlight any new observations or significant changes since the last visit
- For specific queries, prioritize the most recent information but include relevant historical data
- Suggest additional tests or questions based on the available information and any apparent gaps
- If asked about limitations or additional valuable information, be specific about what's missing and how it could improve the assessment
- Always base your answers on the facts provided in the context
- If information to answer a query is not available, clearly state this and explain what information would be needed
- When recommending tests or further questions, explain the rationale based on the patient's history and current condition
- Be prepared to discuss the implications of the MONAI summary in the context of the patient's overall health picture
- If applicable, comment on the consistency or discrepancies between different data sources (e.g., reports vs. observations)
</INSTRUCTIONS>

Patient Context:
{patient_context}

User Query: {user_query}

Please provide a detailed, accurate, and comprehensive answer to the user's query based on the given patient context, following the instructions above.
"""

    assessment_prompt = PromptTemplate.from_template(assessment_prompt_template)

    filled_prompt = assessment_prompt.format(
        patient_context=json.dumps(patient_context, indent=2, default=str),
        user_query=user_query,
    )

    llm = get_llm(
        llm_type=settings.LLM_TYPE,
        temperature=settings.CHAIN_LLM_TEMPERATURE,
    )
    response = llm(filled_prompt)

    return ToolMessage(
        content=response.content.strip(),
        name="AssessmentAssistant",
        tool_call_id="assessment_assistant",
    )


# Function to clear the patient context cache
def clear_patient_context_cache():
    patient_context_cache.clear()
