import logging
from typing import Dict, Any, List, Optional, Tuple
from langchain.prompts import ChatPromptTemplate
from langchain_core.messages import ToolMessage
from langchain.chains import LLMChain
from langchain_community.graphs import Neo4jGraph
from functions.chatbot_api.llm_config import get_llm
from functions.utils import execute_query
import settings
import neo4j
import random
import pandas as pd
import json
from datetime import datetime
from functools import lru_cache

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, pd.Timestamp)):
            return obj.isoformat()
        return super().default(obj)


class ComparisonCache:
    def __init__(self):
        self.cache = {}

    def get(self, condition: str) -> Optional[Dict[str, Any]]:
        if condition in self.cache:
            logger.info(f"✅ Comparison cache hit for condition: {condition}")
            return self.cache[condition]
        logger.info(f"🔎 Comparison cache miss for condition: {condition}")
        return None

    def set(self, condition: str, comparison_data: Dict[str, Any]) -> None:
        logger.info(f"💾 Caching comparison data for condition: {condition}")
        self.cache[condition] = comparison_data

    def clear(self) -> None:
        logger.info("🧹 Clearing comparison cache")
        self.cache.clear()


comparison_cache = ComparisonCache()


class BehavioralChain:
    def __init__(
        self,
        outcomes_fv,
        report_fv,
        patient_data_fv,
        historical_observations_fv,
        report_fg,
        latest_observations_fv,
    ):
        self.outcomes_fv = outcomes_fv
        self.report_fv = report_fv
        self.patient_data_fv = patient_data_fv
        self.historical_observations_fv = historical_observations_fv
        self.report_fg = report_fg
        self.latest_observations_fv = latest_observations_fv
        self.llm = get_llm(
            llm_type=settings.LLM_TYPE,
            temperature=settings.CHAIN_LLM_TEMPERATURE,
        )
        self.query_examples = self.get_query_examples()
        try:
            self.graph = Neo4jGraph(
                settings.NEO4J_URI,
                settings.NEO4J_USER,
                settings.NEO4J_PASSWORD,
                enhanced_schema=True,
            )
        except:
            logger.error(
                "Could not establish connection to Neo4J. Behavioral chain is not going to be available"
            )

    def __call__(self, user_query: str, chat_history: List[Any]) -> ToolMessage:
        query_type = self.classify_query(user_query)

        if query_type == "compare_similar_patients":
            condition = self.extract_condition(user_query)
            if condition:
                return self.handle_compare_similar_patients(user_query, condition)
            else:
                return ToolMessage(
                    content="Could not determine the condition for comparison. Please specify a valid condition.",
                    name="BehavioralQueryAssistant",
                    tool_call_id=1,
                )
        elif query_type == "compare_anomalous_patients":
            condition = self.extract_condition(user_query)
            if condition:
                return self.handle_compare_anomalous_patients(user_query, condition)
            else:
                return ToolMessage(
                    content="Could not determine the condition for comparison. Please specify a valid condition.",
                    name="BehavioralQueryAssistant",
                    tool_call_id=1,
                )
        else:
            return self.handle_graph_query(user_query, query_type)

    def classify_query(self, user_query: str) -> str:
        classification_prompt = ChatPromptTemplate.from_template("""
        Classify the following healthcare-related query into one of these categories:
        1. patient_journey: Queries about a patient's care pathway or journey through the healthcare system
        2. readmission_analysis: Queries about hospital readmissions
        3. condition_query: Queries about specific medical conditions or diagnoses
        4. compare_similar_patients: Compare similar patients with and without a condition (pneumonia, etc)
        5. compare_anomalous_patients: Queries comparing patients with the same condition where the clinical notes are anomalous

        Query: {query}

        Classification:
        """)

        classification_chain = LLMChain(llm=self.llm, prompt=classification_prompt)
        classification = classification_chain.run(query=user_query).strip().lower()

        logger.info(f"⛳️ Behavioral Question Classification: {classification}")

        return classification

    def handle_graph_query(self, user_query: str, query_type: str) -> ToolMessage:
        cypher_generation_prompt = ChatPromptTemplate.from_template("""
        You are an AI assistant trained to generate Cypher queries for a healthcare graph database.

        TASK: Generate a Cypher query to answer the user's question about healthcare data.

        <INSTRUCTIONS>
        - Carefully examine the provided examples and select the ONE most relevant to the user's query.
        - STRICTLY follow the syntax and structure of the chosen example when formulating your response.
        - Use only the provided relationship types and properties in the schema.
        - Do not include any explanations or apologies in your responses.
        - Return your response in the following format:
        <visualization_type> | <generated query>
        - The visualization type should be either 'table' or 'graph'.
        - Ensure the Cypher syntax is correct.
        - DO NOT deviate from the structure of the chosen example unless absolutely necessary to answer the query.
        </INSTRUCTIONS>

        Schema:
        {schema}

        Example Queries:
        {examples}

        User Query: {user_query}
        Query Type: {query_type}

        Response:
        """)

        messages = cypher_generation_prompt.format_prompt(
            schema=self.graph.schema,
            examples=self.query_examples,
            user_query=user_query,
            query_type=query_type,
        ).to_messages()

        response = self.llm(messages)

        # Split the response into visualization type and Cypher query using the | separator
        visualization_type, cypher_query = response.content.strip().split("|", 1)
        visualization_type = visualization_type.strip().lower()
        cypher_query = cypher_query.strip()

        logger.info(f"🪄 Executing query: {cypher_query}")

        transformer = (
            neo4j.Result.to_df if visualization_type == "table" else neo4j.Result.graph
        )
        graph_data = execute_query(cypher_query, transformer=transformer)

        if visualization_type == "table":
            if query_type == "readmission_analysis":
                processed_data = self.process_and_analyze_data(graph_data, user_query)
                summary = "Requested data is retrieved!"
                return ToolMessage(
                    content=processed_data,
                    tool_call_id="behavioral_query",
                    name="BehavioralQueryAssistant",
                )
            else:
                summary = "Requested data is retrieved!"
                return ToolMessage(
                    content=summary,
                    artifact=graph_data,
                    tool_call_id="behavioral_query",
                    name="BehavioralQueryAssistant",
                    response_metadata={"visualization": visualization_type},
                )
        else:
            summary = "I've generated a graph based on your query. The graph shows the relationships you asked about."
            return ToolMessage(
                content=summary,
                artifact=graph_data,
                tool_call_id="behavioral_query",
                name="BehavioralQueryAssistant",
                response_metadata={"visualization": visualization_type},
            )

    def process_and_analyze_data(self, data: pd.DataFrame, user_query: str) -> str:
        # Select the optimal number of observations
        top_observations = data.head(500)

        # Rename columns for clarity
        top_observations.columns = [
            "patient_id",
            "initial_condition",
            "readmission_condition",
        ]

        # Calculate percentages for initial conditions and readmission reasons
        total_readmissions = len(top_observations)
        initial_condition_counts = top_observations["initial_condition"].value_counts()
        initial_condition_percentages = (
            initial_condition_counts / total_readmissions * 100
        ).round(2)

        readmission_counts = top_observations["readmission_condition"].value_counts()
        readmission_percentages = (readmission_counts / total_readmissions * 100).round(
            2
        )

        # Calculate the most common initial condition to readmission condition pairs
        condition_pairs = (
            top_observations.groupby(["initial_condition", "readmission_condition"])
            .size()
            .reset_index(name="count")
        )
        condition_pairs["percentage"] = (
            condition_pairs["count"] / total_readmissions * 100
        ).round(2)
        top_condition_pairs = condition_pairs.sort_values(
            "count", ascending=False
        ).head(10)

        # Prepare the data summaries
        initial_conditions_summary = "\n".join(
            [
                f"{condition}: {percentage}%"
                for condition, percentage in initial_condition_percentages.head(
                    10
                ).items()
            ]
        )
        readmission_reasons_summary = "\n".join(
            [
                f"{reason}: {percentage}%"
                for reason, percentage in readmission_percentages.head(10).items()
            ]
        )
        condition_pairs_summary = "\n".join(
            [
                f"{row['initial_condition']} -> {row['readmission_condition']}: {row['percentage']}%"
                for _, row in top_condition_pairs.iterrows()
            ]
        )

        # Create a prompt for the LLM to analyze the data
        analysis_prompt = ChatPromptTemplate.from_template("""
        Analyze the following summary of patient readmission data:

        Total readmissions analyzed: {total_readmissions}

        Top 10 Initial Conditions:
        {initial_conditions_summary}

        Top 10 Readmission Reasons:
        {readmission_reasons_summary}

        Top 10 Initial Condition to Readmission Condition Pairs:
        {condition_pairs_summary}

        User Query: {user_query}

        Please provide a comprehensive analysis of the data, focusing on:
        1. The most common initial conditions and readmission reasons, with their respective percentages
        2. Any notable patterns or trends in the progression from initial conditions to readmission conditions
        3. Potential implications for patient care and hospital management
        4. Suggestions for reducing readmission rates based on these findings
        5. Any limitations of the data or analysis that should be considered

        Present your analysis in a clear, concise manner using bullet points where appropriate. Use percentages to support your observations.

        Analysis:
        """)

        messages = analysis_prompt.format_prompt(
            total_readmissions=total_readmissions,
            initial_conditions_summary=initial_conditions_summary,
            readmission_reasons_summary=readmission_reasons_summary,
            condition_pairs_summary=condition_pairs_summary,
            user_query=user_query,
        ).to_messages()

        # Get the LLM's analysis
        analysis = self.llm(messages)

        return analysis.content.strip()

    def handle_compare_similar_patients(
        self, user_query: str, condition: str
    ) -> ToolMessage:
        comparative_data = self.get_comparative_patient_data(condition)

        prompt_template = ChatPromptTemplate.from_template("""
        Compare patients with {condition} to similar patients without the condition.

        Patients with {condition}:
        {patients_with_condition}

        Similar patients without {condition}:
        {patients_without_condition}

        Provide a detailed analysis focusing on:
        1. Key similarities and differences between the two groups
        2. Potential impact of {condition} on patient health and prognosis
        3. Any notable patterns or trends in the patient data
        4. Implications for treatment and care strategies

        User Query: {user_query}

        Analysis:
        """)

        messages = prompt_template.format_prompt(
            condition=condition,
            patients_with_condition=comparative_data["patients_with_condition"],
            patients_without_condition=comparative_data["patients_without_condition"],
            user_query=user_query,
        ).to_messages()

        response = self.llm(messages)

        return ToolMessage(
            content=response.content.strip(),
            name="BehavioralQueryAssistant",
            tool_call_id=2,
        )

    def handle_compare_anomalous_patients(
        self, user_query: str, condition: str
    ) -> ToolMessage:
        # Check cache first
        cached_comparison = comparison_cache.get(condition)
        if cached_comparison:
            comparison_result = cached_comparison
        else:
            closest_patient, furthest_patient = self.find_comparison_patients(condition)
            comparison_result = self.compare_patients(closest_patient, furthest_patient)
            # Cache the comparison result
            comparison_cache.set(condition, comparison_result)

        prompt_template = ChatPromptTemplate.from_template("""
        Compare the following two patients with {condition}. One patient represents a typical case, while the other presents anomalous characteristics:

        {comparison_result}

        Provide a detailed analysis of the differences between these patients, focusing on:

        1. Introduction: Briefly summarize the key characteristics of both patients and the purpose of the comparison.

        2. Key Differences in Their Medical History:
           - Age and Demographics
           - Relevant Medical History
           - Lifestyle Factors
           - Socioeconomic Background
           - Current Medications
           - Recent Reports and Symptoms

        3. Anomalies in the Clinical Notes of the Anomalous Patient:
           - Unusual symptoms or lack thereof
           - Differences in treatment approaches
           - Inconsistencies or gaps in documentation
           - Any other atypical features

        4. Potential Implications of These Differences for Diagnosis and Treatment:
           - How the differences might affect the diagnosis of {condition}
           - Implications for treatment plans and their effectiveness
           - Potential risks or complications for each patient
           - Suggestions for further tests or monitoring

        5. Conclusion: Summarize the key points of the comparison and their significance for patient care in cases of {condition}.

        In your analysis:
        - Refer to the patients as "Typical Patient" and "Anomalous Patient"
        - Ensure equal attention is given to both patients throughout the analysis
        - Be specific in relating observed differences to the {condition}
        - Use clear formatting with headers and bullet points for readability

        User Query: {user_query}

        Analysis:
        """)

        messages = prompt_template.format_prompt(
            condition=condition,
            comparison_result=json.dumps(comparison_result, indent=2, cls=JSONEncoder),
            user_query=user_query,
        ).to_messages()

        response = self.llm(messages)

        return ToolMessage(
            content=response.content.strip(),
            name="BehavioralQueryAssistant",
            tool_call_id=3,
        )

    @lru_cache(maxsize=100)
    def extract_condition(self, query: str) -> Optional[str]:
        conditions = [
            "atelectasis",
            "cardiomegaly",
            "consolidation",
            "edema",
            "enlarged_cardiomediastinum",
            "fracture",
            "lung_lesion",
            "lung_opacity",
            "no_finding",
            "pleural_effusion",
            "pleural_other",
            "pneumonia",
            "pneumothorax",
            "support_devices",
        ]

        prompt_template = ChatPromptTemplate.from_template("""
        Extract one medical condition from the following query. The condition should be one of: {conditions}

        Query: {query}

        If no specific condition is mentioned, respond with "unknown".

        Extracted condition:
        """)

        extraction_chain = LLMChain(llm=self.llm, prompt=prompt_template)
        extracted_condition = (
            extraction_chain.run(conditions=", ".join(conditions), query=query)
            .strip()
            .lower()
        )

        # Remove any "extracted condition:" prefix if present
        extracted_condition = extracted_condition.replace(
            "extracted condition:", ""
        ).strip()

        logger.info(f"🧪 Extracted condition: {extracted_condition}")
        return extracted_condition if extracted_condition in conditions else None

    def find_comparison_patients(self, condition: str) -> Tuple[int, int]:
        outcomes_df = self.outcomes_fv.get_batch_data()
        patient_ids = outcomes_df[outcomes_df[condition] == 1].patient_id.values

        report_embeddings = self.report_fv.get_batch_data()
        report_embeddings_filtered = report_embeddings[
            report_embeddings["patient_id"].isin(patient_ids)
        ]

        typical_patient_embedding = report_embeddings_filtered["embeddings"].mean()

        id_filter = self.report_fg.patient_id.contains(patient_ids.tolist())
        id_filter._value = patient_ids.tolist()

        closest_patient_id = self.report_fv.find_neighbors(
            typical_patient_embedding, k=1, filter=id_filter
        )
        if not closest_patient_id:
            raise ValueError(f"No patients found for condition: {condition}")
        closest_patient_id = closest_patient_id[0][0]

        furthest_patient_id = self.report_fv.find_neighbors(
            typical_patient_embedding, k=len(patient_ids), filter=id_filter
        )[-1][0]

        return closest_patient_id, furthest_patient_id

    def compare_patients(
        self, closest_patient_id: int, furthest_patient_id: int
    ) -> Dict[str, Any]:
        patient_data = self.patient_data_fv.get_batch_data()

        closest_patient_data = (
            patient_data[patient_data["patient_id"] == closest_patient_id]
            .sort_values("report_date", ascending=False)
            .head(3)
        )
        furthest_patient_data = (
            patient_data[patient_data["patient_id"] == furthest_patient_id]
            .sort_values("report_date", ascending=False)
            .head(3)
        )

        historical_observations_data = self.historical_observations_fv.get_batch_data()
        closest_patient_historical = historical_observations_data[
            historical_observations_data["patient_id"] == closest_patient_id
        ].head()
        furthest_patient_historical = historical_observations_data[
            historical_observations_data["patient_id"] == furthest_patient_id
        ].head()

        closest_patient_latest = self.latest_observations_fv.get_feature_vector(
            {"patient_id": closest_patient_id}
        )
        furthest_patient_latest = self.latest_observations_fv.get_feature_vector(
            {"patient_id": furthest_patient_id}
        )

        comparison_result = {
            "closest_patient": {
                "id": int(closest_patient_id),
                "recent_reports": closest_patient_data.to_dict("records"),
                "historical_observations": closest_patient_historical.to_dict(
                    "records"
                ),
                "latest_observation": closest_patient_latest,
            },
            "furthest_patient": {
                "id": int(furthest_patient_id),
                "recent_reports": furthest_patient_data.to_dict("records"),
                "historical_observations": furthest_patient_historical.to_dict(
                    "records"
                ),
                "latest_observation": furthest_patient_latest,
            },
        }

        return json.loads(json.dumps(comparison_result, cls=JSONEncoder))

    def get_comparative_patient_data(self, condition: str) -> Dict[str, str]:
        patient_data = self.patient_data_fv.get_batch_data()
        historical_observations_data = self.historical_observations_fv.get_batch_data()

        patients_with_condition = patient_data[patient_data[condition] == 1]
        patients_with_condition_ids = random.sample(
            list(patients_with_condition["patient_id"].values), 3
        )

        report_data = self.report_fv.get_batch_data()
        patient_embeddings = (
            report_data[report_data["patient_id"].isin(patients_with_condition_ids)]
            .groupby("patient_id")
            .mean()
        )

        patients_without_condition = patient_data[patient_data["no_finding"] == 1]
        patients_without_condition_ids = patients_without_condition["patient_id"].values

        id_filter = self.report_fg.patient_id.contains(
            [int(pid) for pid in patients_without_condition_ids]
        )
        id_filter._value = [int(pid) for pid in patients_without_condition_ids]

        similar_patients_without_condition_ids = [
            self.report_fv.find_neighbors(embedding, k=1, filter=id_filter)[0][0]
            for embedding in patient_embeddings["embeddings"]
        ]

        patients_with_condition_info = self.get_patient_group_information(
            patients_with_condition_ids,
            condition,
            patient_data,
            historical_observations_data,
        )
        patients_without_condition_info = self.get_patient_group_information(
            similar_patients_without_condition_ids,
            "NO FINDINGS",
            patient_data,
            historical_observations_data,
        )

        return {
            "patients_with_condition": patients_with_condition_info,
            "patients_without_condition": patients_without_condition_info,
        }

    def get_patient_group_information(
        self, patient_ids, condition_value, patient_data, historical_observations_data
    ):
        patient_information_list = []
        for patient_id in patient_ids:
            patient_information = self.get_patient_information(
                patient_id, patient_data, historical_observations_data
            )
            patient_information_list.append(patient_information)
        patient_information_formatted = (
            f"PATIENTS WITH {condition_value}:\n\n"
            + "\n\n".join(patient_information_list)
        )
        return patient_information_formatted

    def get_patient_information(
        self, patient_id, patient_data, historical_observations_data
    ):
        patient_report = (
            patient_data[patient_data["patient_id"] == patient_id]
            .sort_values("report_date", ascending=False)["report_text"]
            .iloc[0]
        )

        patient_historical_observations = (
            historical_observations_data[
                historical_observations_data["patient_id"] == patient_id
            ]
            .sort_values("observation_date", ascending=False)[
                ["observation_date", "observations"]
            ]
            .head(3)
        )

        patient_historical_observations_formatted = (
            patient_historical_observations.apply(self.format_observations, axis=1)
        )
        patient_information = (
            "PATIENT REPORT:\n"
            + patient_report
            + "\n\nHISTORICAL OBSERVATIONS:\n"
            + "\n".join(patient_historical_observations_formatted)
        )

        return patient_information

    @staticmethod
    def format_observations(row):
        date = row["observation_date"].strftime("%Y-%m-%d %H:%M:%S UTC")
        obs = row["observations"].split("|")
        formatted_obs = "\n".join(f"  - {o.strip()}" for o in obs)
        return f"## Observation\n- Date: {date}\n- Observations:\n{formatted_obs}\n"

    def get_query_examples(self) -> str:
        return """
        1. Patient Journey Query:
        User Query: "Can you show me the journey of a patient who has been diagnosed with acute bronchitis?"
        graph | MATCH(p:Patient)-[:HAS_ENCOUNTER]-(e:Encounter)-[:HAS_CONDITION]-(c:Condition)
        WHERE c.description="Acute bronchitis (disorder)"
        WITH p LIMIT 1
        MATCH journey=((p:Patient)-[:HAS_ENCOUNTER]->(e1:Encounter)-[:NEXT*]->())
        WHERE count{()-[:NEXT]->(e1)} = 0
        RETURN journey

        2. Readmission Analysis Query:
        User Query: "What are the most common reasons patients get readmitted to the hospital within 30 days?"
        table | MATCH(e1:Encounter)<-[:HAS_ENCOUNTER]-(p:Patient)-[:HAS_ENCOUNTER]->(e2:Encounter)
        WHERE e1.class='inpatient' and e2.class='inpatient'
        AND abs(duration.between(e1.date, e2.date).milliseconds) < 2592000000
        WITH p, e1, e2
        MATCH (c1:Condition)<-[:HAS_CONDITION]-(e1)<-[:HAS_ENCOUNTER]-(p)-[:HAS_ENCOUNTER]->(e2)-[:HAS_CONDITION]->(c2:Condition)
        RETURN p.id, c1.description, c2.description

        3. Condition Query:
        User Query: "Can you show me patients with condition code 444814009?"
        table | MATCH path=(patient:Patient)-[:HAS_ENCOUNTER]->(encounter)-[:HAS_CONDITION]->(condition {code:"444814009"})
        RETURN patient.SSN as `Patient ID`, condition.code as `Condition Code`, condition.description as `Description`

        4. Pathway for a patient who has been readmitted:
        User Query: "I'd like to see the care pathway for a patient who has been readmitted to the hospital. Can you show me that?"
        graph | MATCH(e1:Encounter)<-[:HAS_ENCOUNTER]-(p:Patient)-[:HAS_ENCOUNTER]->(e2:Encounter)
        WHERE e1.class='inpatient' and e2.class='inpatient'
        AND abs(duration.between(e1.date, e2.date).milliseconds) < 2592000000
        WITH p LIMIT 1
        MATCH journey= ((p:Patient)-[:HAS_ENCOUNTER]->(e:Encounter)-[:NEXT]->())
        RETURN journey
        """


# Function to clear the comparison cache
def clear_comparison_cache():
    comparison_cache.clear()
