import logging
from typing import Dict, Any, Optional
from langchain.prompts import PromptTemplate
from langchain_core.messages import ToolMessage
from langchain.chains import LLMChain
from functions.chatbot_api.llm_config import get_llm
import settings

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class DetectionChain:
    def __init__(
        self,
        report_fg,
        outcomes_fv,
        report_fv,
        patient_data_fv,
        historical_observations_fv,
        latest_observations_fv,
    ):
        self.report_fg = report_fg
        self.outcomes_fv = outcomes_fv
        self.report_fv = report_fv
        self.patient_data_fv = patient_data_fv
        self.historical_observations_fv = historical_observations_fv
        self.latest_observations_fv = latest_observations_fv
        self.llm = get_llm(
            llm_type=settings.LLM_TYPE,
            temperature=settings.CHAIN_LLM_TEMPERATURE,
        )

    def __call__(
        self, user_query: str, monai_detection: str = "", report_text: str = ""
    ) -> ToolMessage:
        query_type = self.classify_query(user_query)

        if query_type == "comprehensive_analysis":
            return self.handle_comprehensive_analysis(monai_detection, report_text)
        elif query_type == "summary_classification":
            return self.handle_summary_classification(
                user_query, monai_detection, report_text
            )
        elif query_type == "observation_interpretation":
            return self.handle_observation_interpretation(user_query)
        else:
            return ToolMessage(
                content="I'm not sure how to handle this type of query.",
                name="DetectionAssistant",
                tool_call_id=1,
            )

    def classify_query(self, user_query: str) -> str:
        classification_prompt = PromptTemplate.from_template("""
        Classify the following query into one of these categories:
        1. comprehensive_analysis: Requests for a comprehensive, data-driven analysis of MONAI predictions and radiologist's report
        2. summary_classification: Queries about image summary or classification
        3. observation_interpretation: Queries about interpreting specific observations

        Please return only the classification name either: comprehensive_analysis, summary_classification or observation_interpretation

        Query: {query}

        Classification:
        """)

        classification_chain = LLMChain(llm=self.llm, prompt=classification_prompt)
        classification = classification_chain.run(query=user_query).strip().lower()

        logger.info(f"⛳️ Question Classification: {classification}")

        return classification

    def handle_comprehensive_analysis(
        self, monai_detection: str, report_text: str
    ) -> ToolMessage:
        prompt_template = """
        You are an AI assistant specializing in analyzing medical imaging reports and model predictions. 
        Based on the provided MONAI model predictions and the radiologist's report, provide a comprehensive, data-driven analysis of the findings.

        MONAI Model Predictions:
        {monai_detection}

        Radiologist's Report:
        {report_text}

        Please provide a detailed analysis covering the following points:

        1. Model Predictions Analysis:
        - List the top 5 highest probability predictions with their exact numerical values.
        - For each of these top 5 predictions, provide a brief explanation of what they indicate clinically.
        - Identify any low probability predictions (below 10%) that might still be clinically relevant and explain why.

        2. Radiologist's Report Analysis:
        - List the key findings mentioned in the report using bullet points.
        - Highlight any specific measurements or numerical data provided in the report.

        3. Correlation Between Model Predictions and Radiologist's Report:
        - For each of the top 5 model predictions, state whether it is confirmed, partially confirmed, or contradicted by the radiologist's report. Provide specific quotes from the report to support your assessment.
        - Identify any significant findings in the radiologist's report that were not highlighted by the model predictions.

        4. Discrepancies and Explanations:
        - List any clear discrepancies between the model predictions and the radiologist's report.
        - For each discrepancy, provide possible explanations for why the model and radiologist might disagree.

        5. Summary of Key Points:
        - Summarize the most important findings, integrating both the model predictions and the radiologist's observations.
        - Highlight any areas where further investigation might be needed based on the combined analysis.

        Ensure your analysis is concise, focuses on numerical data where available, and provides clear correlations between the model predictions and the radiologist's observations.
        """

        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(monai_detection=monai_detection, report_text=report_text)
        )

        return ToolMessage(
            content=response.content.strip(), name="DetectionAssistant", tool_call_id=2
        )

    def handle_summary_classification(
        self, user_query: str, monai_detection: str, report_text: str
    ) -> ToolMessage:
        prompt_template = """
        You are an advanced AI trained to understand medical records.
        TASK: Summarize the different medical records and reports presented to you. 

        <INSTRUCTIONS>
        - The answer must contain 4 sections: Summary of the findings, summary of impressions, model predictions and conclusions 
        - You should always include the 4 sections above and no additional section should be included.
        - If findings or impressions are not available in the report, you should leave the section empty.
        - You should summarize the model predictions and put them in context with the findings and impressions
        - The section header should be a markdown header with 3 #
        - You should highlight in markdown bold the information you think are most relevant to the case
        </INSTRUCTIONS>

        Doctor's question: {question}
        MONAI Detection: {monai_detection}
        Report text: {report_text}

        Response:
        """

        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(
                question=user_query,
                monai_detection=monai_detection,
                report_text=report_text,
            )
        )

        return ToolMessage(
            content=response.content.strip(), name="DetectionAssistant", tool_call_id=3
        )

    def handle_observation_interpretation(self, user_query: str) -> ToolMessage:
        prompt_template = """
        Interpret the following observation in the context of a medical test:

        Observation: {query}

        Provide an explanation of:
        1. What this test measures
        2. The normal range for this measurement
        3. Potential implications of results outside the normal range
        4. Any additional tests or follow-ups that might be recommended based on this result

        Interpretation:
        """

        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(prompt.format(query=user_query))

        return ToolMessage(
            content=response.content.strip(), name="DetectionAssistant", tool_call_id=4
        )
