import logging
from typing import Dict, Any, Optional
from langchain.prompts import PromptTemplate
from langchain_core.messages import ToolMessage
import settings
from functions.chatbot_api.llm_config import get_llm
import json
import pandas as pd
from datetime import datetime

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class PatientDiagnosisCache:
    def __init__(self):
        self.cache = {}

    def get(self, patient_id: int) -> Optional[Dict[str, Any]]:
        if patient_id in self.cache:
            logger.info(f"✅ Cache hit for patient ID: {patient_id}")
            return self.cache[patient_id]
        logger.info(f"🔎 Cache miss for patient ID: {patient_id}")
        return None

    def set(self, patient_id: int, context: Dict[str, Any]) -> None:
        logger.info(f"💾 Caching context for patient ID: {patient_id}")
        self.cache[patient_id] = context

    def clear(self) -> None:
        logger.info("🧹 Clearing patient diagnosis cache")
        self.cache.clear()


patient_diagnosis_cache = PatientDiagnosisCache()


def get_patient_diagnosis_context(
    patient_id: int,
    historical_observations_fv: Any,
    latest_observations_fv: Any,
    monai_detection: str,
) -> Dict[str, Any]:
    # Retrieve cached patient context if present
    cached_context = patient_diagnosis_cache.get(patient_id)
    if cached_context:
        cached_context["monai_detection"] = monai_detection
        return cached_context

    logger.info(
        f"🪝 Retrieving patient diagnosis context for ID from the Feature Store: {patient_id}"
    )

    # Retrieve historical observations
    historical_observations = historical_observations_fv.get_batch_data()
    patient_observations = historical_observations[
        historical_observations.patient_id == patient_id
    ]
    patient_observations_latest10 = patient_observations.sort_values(
        "observation_date", ascending=False
    ).head(10)

    # Get latest new observation
    try:
        latest_observation = latest_observations_fv.get_feature_vector(
            {"patient_id": patient_id}
        )
    except Exception as e:
        logger.error(f"Error retrieving latest observation: {str(e)}")
        latest_observation = None

    # Prepare patient context
    context = prepare_diagnosis_context(
        patient_observations_latest10, latest_observation, monai_detection
    )

    # Cache the patient context
    patient_diagnosis_cache.set(patient_id, context)

    return context


def prepare_diagnosis_context(
    patient_observations_latest10, latest_observation, monai_detection
):
    def format_observations(obs_string: str) -> list:
        return [o.strip() for o in obs_string.split("|")]

    def format_date(date):
        if isinstance(date, pd.Timestamp):
            return date.isoformat()
        elif isinstance(date, datetime):
            return date.isoformat()
        return str(date)

    print("================================================")
    print(latest_observation)
    print("================================================")

    context = {
        "historical_observations": [
            {
                "date": format_date(row["observation_date"]),
                "observations": format_observations(str(row["observations"])),
            }
            for _, row in patient_observations_latest10.iterrows()
        ],
        "latest_observation_since_last_visit": {
            "date": format_date(latest_observation[1])
            if latest_observation
            else "No date available",
            "observations": format_observations(str(latest_observation[2]))
            if latest_observation
            else ["No new observations"],
        }
        if latest_observation
        else {"date": "No date available", "observations": ["No new observations"]},
        "monai_detection": monai_detection,
    }

    return context


def diagnosis_assistant(
    user_query: str,
    patient_id: int,
    historical_observations_fv: Any,
    latest_observations_fv: Any,
    monai_detection: str,
) -> ToolMessage:
    """Function to answer diagnosis-related queries based on patient context."""
    logger.info(f"👨🏻‍⚕️ Diagnosis query for patient ID: {patient_id}")

    patient_context = get_patient_diagnosis_context(
        patient_id,
        historical_observations_fv,
        latest_observations_fv,
        monai_detection,
    )

    diagnosis_prompt_template = """
You are an advanced AI trained to support doctors in diagnosing patient conditions based on historical observations, recent data, and MONAI detection results.
TASK: Answer the user's query based on the provided patient context and MONAI detection summary.

<INSTRUCTIONS>
- Utilize ALL provided patient context and MONAI detection summary to answer queries comprehensively, including:
  * Historical observations (up to 10 most recent)
  * Latest observation since the last visit
  * MONAI detection results
- For queries about available observations, provide a clear summary of what data is available
- When assessing if there's enough information for a diagnosis:
  * Analyze the available data, MONAI detection results, and their relevance to potential conditions
  * If a diagnosis can be made, provide it with supporting evidence
  * If more information is needed, specify what additional data or tests would be helpful
- When suggesting additional questions for the patient:
  * Base suggestions on gaps in the current data and MONAI detection results
  * Explain why each suggested question is important for diagnosis
- Always base your answers on the facts provided in the context and MONAI detection summary
- If information to answer a query is not available, clearly state this and explain what information would be needed
- Be prepared to discuss trends or changes in the patient's condition over time, considering the MONAI detection results
- If applicable, comment on the consistency or discrepancies between different observations and the MONAI detection results
</INSTRUCTIONS>

Patient Context:
{patient_context}
d
MONAI Detection:
{monai_detection}

User Query: {user_query}

Please provide a detailed, accurate, and comprehensive answer to the user's query based on the given patient context and MONAI detection summary, following the instructions above.
"""

    diagnosis_prompt = PromptTemplate.from_template(diagnosis_prompt_template)

    filled_prompt = diagnosis_prompt.format(
        patient_context=json.dumps(patient_context, indent=2, default=str),
        monai_detection=monai_detection,
        user_query=user_query,
    )

    llm = get_llm(
        llm_type=settings.LLM_TYPE,
        temperature=settings.CHAIN_LLM_TEMPERATURE,
    )
    response = llm(filled_prompt)

    return ToolMessage(
        content=response.content.strip(),
        name="DiagnosisAssistant",
        tool_call_id="diagnosis_assistant",
    )


# Function to clear the patient diagnosis cache
def clear_patient_diagnosis_cache():
    patient_diagnosis_cache.clear()
