import logging
from typing import Dict, Any, List, Optional
from langchain.prompts import PromptTemplate
from langchain_core.messages import ToolMessage, AIMessage
from langchain.chains import LLMChain
from functions.chatbot_api.llm_config import get_llm
import settings
import json
import numpy as np
import pandas as pd

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class PrognosisCache:
    def __init__(self):
        self.cache = {}

    def get(self, patient_ssn: str) -> Optional[str]:
        if patient_ssn in self.cache:
            logger.info(f"✅ Prognosis cache hit for patient SSN: {patient_ssn}")
            return self.cache[patient_ssn]
        logger.info(f"🔎 Prognosis cache miss for patient SSN: {patient_ssn}")
        return None

    def set(self, patient_ssn: str, prognosis: str) -> None:
        logger.info(f"💾 Caching prognosis for patient SSN: {patient_ssn}")
        self.cache[patient_ssn] = prognosis

    def clear(self) -> None:
        logger.info("🧹 Clearing prognosis cache")
        self.cache.clear()


prognosis_cache = PrognosisCache()


class PrognosisChain:
    def __init__(self, outcomes_fv, report_fv, prognoses_fv, report_fg):
        self.outcomes_fv = outcomes_fv
        self.report_fv = report_fv
        self.prognoses_fv = prognoses_fv
        self.report_fg = report_fg
        self.llm = get_llm(
            llm_type=settings.LLM_TYPE,
            temperature=settings.CHAIN_LLM_TEMPERATURE,
        )

    def __call__(
        self,
        user_query: str,
        patient_ssn: str,
        chat_history: List[Any],
        monai_detection: str,
    ) -> ToolMessage:
        query_type = self.classify_query(user_query)

        if query_type == "general_prognosis":
            response = self.handle_general_prognosis(
                user_query, patient_ssn, chat_history, monai_detection
            )
            self.cache_prognosis(patient_ssn, response.content)
            return response
        elif query_type == "comparative_prognosis":
            return self.handle_comparative_prognosis(
                user_query, patient_ssn, chat_history
            )
        elif query_type == "prognosis_information":
            return self.handle_prognosis_information(
                user_query, patient_ssn, chat_history, monai_detection
            )
        else:
            return ToolMessage(
                content="I'm not sure how to handle this type of prognosis query.",
                name="PrognosisAssistant",
                tool_call_id=1,
            )

    def cache_prognosis(self, patient_ssn: str, prognosis: str):
        prognosis_cache.set(patient_ssn, prognosis)

    def classify_query(self, user_query: str) -> str:
        classification_prompt = PromptTemplate.from_template("""
        Classify the following prognosis-related query into one of these categories:
        1. general_prognosis: Queries about making a prognosis based on available information
        2. comparative_prognosis: Queries comparing the prognosis of a patient with similar patients
        3. prognosis_information: Queries about additional information needed for a more accurate prognosis

        Please return only the classification name either: general_prognosis, comparative_prognosis or prognosis_information 

        Query: {query}

        Classification:
        """)

        classification_chain = LLMChain(llm=self.llm, prompt=classification_prompt)
        classification = classification_chain.run(query=user_query).strip().lower()

        logger.info(f"⛳️ Prognosis Question Classification: {classification}")

        return classification

    def handle_general_prognosis(
        self,
        user_query: str,
        patient_ssn: str,
        chat_history: List[Any],
        monai_detection: str,
    ) -> ToolMessage:
        prompt_template = """
        You are an advanced AI trained to support doctors in making prognoses based on patient information and MONAI detection results.
        TASK: Answer the user's query about patient prognosis based on the available patient information and MONAI detection summary.

        <INSTRUCTIONS>
        - Analyze the available patient information and MONAI detection results to gather all relevant data for making a prognosis.
        - Assess the completeness and relevance of the available patient information and MONAI detection results.
        - If there's enough information, provide a detailed prognosis with supporting evidence, incorporating insights from the MONAI detection.
        - If more information is needed, specify what additional data or tests would be helpful.
        - Always base your answers on the facts provided in the patient information and MONAI detection summary.
        - If critical information for a prognosis is not available, clearly state this and explain what information would be needed.
        - Be prepared to discuss how different factors in the patient's history and the MONAI detection results might affect the prognosis.
        - If applicable, comment on the reliability of the prognosis based on the available information and MONAI detection results.
        - Avoid making definitive statements about the patient's future if the information is insufficient.
        - When appropriate, suggest follow-up assessments or monitoring to refine the prognosis over time.
        </INSTRUCTIONS>

        Patient information:
        {patient_info}

        MONAI Detection:
        {monai_detection}

        User Query: {user_query}

        Please provide a detailed, accurate, and comprehensive answer to the user's query about prognosis based on the given patient information and MONAI detection summary, following the instructions above.
        """

        patient_info = self.get_patient_info(patient_ssn, chat_history)
        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(
                user_query=user_query,
                patient_info=patient_info,
                monai_detection=monai_detection,
            )
        )

        return ToolMessage(
            content=response.content.strip(), name="PrognosisAssistant", tool_call_id=2
        )

    def handle_comparative_prognosis(
        self, user_query: str, patient_ssn: str, chat_history: List[Any]
    ) -> ToolMessage:
        condition = self.extract_condition(chat_history)
        if not condition:
            return ToolMessage(
                content="Could not determine the patient's condition for comparison. Please specify a valid condition.",
                name="PrognosisAssistant",
                tool_call_id=3,
            )

        comparative_data = self.get_comparative_prognosis(patient_ssn, condition)

        prompt_template = """
        Compare the prognosis of the current patient with similar patients who have {condition}.

        Current Patient Prognosis:
        {current_prognosis}

        Similar Patients' Prognoses:
        {comparative_prognoses}

        Provide a detailed analysis focusing on:
        1. Key similarities and differences in prognoses
        2. Factors that might account for these differences
        3. Implications for the current patient's treatment and care
        4. Potential outcomes based on the comparative analysis

        User Query: {user_query}

        Analysis:
        """

        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(
                condition=condition,
                current_prognosis=comparative_data["current_patient_prognosis"],
                comparative_prognoses=json.dumps(
                    comparative_data["similar_patients_prognoses"], indent=2
                ),
                user_query=user_query,
            )
        )

        return ToolMessage(
            content=response.content.strip(), name="PrognosisAssistant", tool_call_id=4
        )

    def handle_prognosis_information(
        self,
        user_query: str,
        patient_ssn: str,
        chat_history: List[Any],
        monai_detection,
    ) -> ToolMessage:
        prompt_template = """
        Based on the provided patient information and the user's query about additional information needed for a more accurate prognosis, please:

        1. Identify any gaps in the current patient information that could affect the accuracy of the prognosis.
        2. Suggest specific questions to ask the patient or additional tests to perform.
        3. Explain why each suggested question or test is important for refining the prognosis.
        4. If applicable, prioritize the suggested questions or tests based on their potential impact on the prognosis.

        Patient Information:
        {patient_info}

        MonAI Detection:
        {monai_detection}

        User Query: {user_query}

        Response:
        """

        patient_info = self.get_patient_info(patient_ssn, chat_history)
        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(
                patient_info=patient_info,
                user_query=user_query,
                monai_detection=monai_detection,
            )
        )

        return ToolMessage(
            content=response.content.strip(), name="PrognosisAssistant", tool_call_id=5
        )

    def extract_condition(self, chat_history: List[Any]) -> Optional[str]:
        conditions = [
            "lung_opacity",
            "atelectasis",
            "cardiomegaly",
            "consolidation",
            "edema",
            "enlarged_cardiomediastinum",
            "fracture",
            "lung_lesion",
            "no_finding",
            "pleural_effusion",
            "pleural_other",
            "pneumonia",
            "pneumothorax",
            "support_devices",
        ]

        chat_text = " ".join(
            [msg.content for msg in chat_history if isinstance(msg, AIMessage)]
        )

        for condition in conditions:
            if condition in chat_text.lower():
                return condition

        return None

    def get_patient_info(self, patient_ssn: str, chat_history: List[Any]) -> str:
        return json.dumps(
            [msg.content for msg in chat_history if isinstance(msg, AIMessage)],
            indent=2,
        )

    def get_comparative_prognosis(
        self, patient_ssn: str, condition: str
    ) -> Dict[str, Any]:
        cached_prognosis = prognosis_cache.get(patient_ssn)

        outcomes_df = self.outcomes_fv.get_batch_data()
        similar_patient_ids = outcomes_df[
            (outcomes_df[condition] == 1) & (outcomes_df["patient_id"] != patient_ssn)
        ].patient_id.values

        report_embeddings = self.report_fv.get_batch_data()
        patient_embedding_df = report_embeddings[
            report_embeddings["patient_id"] == patient_ssn
        ]
        patient_embedding = np.mean(patient_embedding_df["embeddings"].tolist(), axis=0)

        similar_patient_ids_list = [str(pid) for pid in similar_patient_ids]
        id_filter = self.report_fg.patient_id.contains(similar_patient_ids_list)
        id_filter._value = similar_patient_ids_list

        closest_patients = self.report_fv.find_neighbors(
            patient_embedding, k=2, filter=id_filter
        )
        closest_patients_id = [patient[0] for patient in closest_patients]

        if len(closest_patients_id) != 0:
            closest_patient_prognoses = self.prognoses_fv.get_feature_vectors(
                [{"patient_id": pid} for pid in closest_patients_id]
            )
        else:
            closest_patient_prognoses = []

        comparative_data = {
            "condition": condition,
            "current_patient_prognosis": cached_prognosis
            if cached_prognosis
            else "No cached prognosis available",
            "similar_patients_prognoses": closest_patient_prognoses,
        }

        return comparative_data


def clear_prognosis_cache():
    prognosis_cache.clear()
