import logging
from typing import Dict, Any, List, Optional
from langchain.prompts import PromptTemplate
from langchain_core.messages import ToolMessage
from langchain.prompts import ChatPromptTemplate
from langchain.chains import LLMChain
from functions.chatbot_api.llm_config import get_llm
import settings
import numpy as np
from functools import lru_cache

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class TreatmentPlanCache:
    def __init__(self):
        self.cache = {}

    def get(self, patient_ssn: str) -> Optional[str]:
        if patient_ssn in self.cache:
            logger.info(f"✅ Treatment plan cache hit for patient SSN: {patient_ssn}")
            return self.cache[patient_ssn]
        logger.info(f"🔎 Treatment plan cache miss for patient SSN: {patient_ssn}")
        return None

    def set(self, patient_ssn: str, treatment_plan: str) -> None:
        logger.info(f"💾 Caching treatment plan for patient SSN: {patient_ssn}")
        self.cache[patient_ssn] = treatment_plan

    def clear(self) -> None:
        logger.info("🧹 Clearing treatment plan cache")
        self.cache.clear()


treatment_plan_cache = TreatmentPlanCache()


class TreatmentChain:
    def __init__(self, outcomes_fv, treatment_plans_fv):
        self.outcomes_fv = outcomes_fv
        self.treatment_plans_fv = treatment_plans_fv
        self.llm = get_llm(
            llm_type=settings.LLM_TYPE,
            temperature=settings.CHAIN_LLM_TEMPERATURE,
        )

    def __call__(
        self,
        user_query: str,
        monai_detection: str,
        patient_ssn: str,
        chat_history: List[Any],
    ) -> ToolMessage:
        query_type = self.classify_query(user_query)

        if query_type == "check_existing_plan":
            response = self.handle_check_existing_plan(patient_ssn)
        elif query_type == "create_new_plan":
            response = self.handle_create_new_plan(
                monai_detection, patient_ssn, chat_history
            )
        elif query_type == "update_existing_plan":
            response = self.handle_update_existing_plan(
                monai_detection, patient_ssn, chat_history
            )
        elif query_type == "outline_plan_timeline":
            response = self.handle_outline_plan_timeline(patient_ssn, chat_history)
        elif query_type == "next_immediate_step":
            response = self.handle_next_immediate_step(
                monai_detection, patient_ssn, chat_history
            )
        elif query_type == "example_treatment_plans":
            response = self.handle_example_treatment_plans(monai_detection)
        elif query_type == "compare_treatment_plans":
            response = self.handle_compare_treatment_plans(patient_ssn, chat_history)
        else:
            response = "I'm not sure how to handle this type of treatment plan query. Please rephrase your question."

        return ToolMessage(
            content=response, name="TreatmentAssistant", tool_call_id="1"
        )

    def classify_query(self, user_query: str) -> str:
        classification_prompt = PromptTemplate.from_template("""
        Classify the following treatment-related query into one of these categories (Do NOT include the number):
        1. check_existing_plan: Queries about whether a patient has an existing treatment plan
        2. create_new_plan: Requests to create a new comprehensive treatment plan
        3. update_existing_plan: Requests to update an existing plan based on changes in the patient's condition
        4. outline_plan_timeline: Requests to outline the treatment plan for specific time periods
        5. next_immediate_step: Queries about the next immediate step in the patient's care
        6. example_treatment_plans: Requests for example treatment plans for patients with a particular condition
        7. compare_treatment_plans: Requests to compare the patient's prior treatment plan with expected treatment plan

        Please return only return a value from the following list and no other text: 
        - check_existing_plan
        - create_new_plan
        - update_existing_plan
        - outline_plan_timeline
        - next_immediate_step
        - example_treatment_plans
        - compare_treatment_plans

        Only return

        Query: {query}

        Classification:
        """)

        classification_chain = LLMChain(llm=self.llm, prompt=classification_prompt)
        classification = classification_chain.run(query=user_query).strip().lower()

        logger.info(f"⛳️ Treatment Question Classification: {classification}")

        return classification

    def handle_check_existing_plan(self, patient_ssn: str) -> str:
        cached_plan = treatment_plan_cache.get(patient_ssn)
        if cached_plan:
            return f"Yes, the patient has an existing treatment plan:\n\n{cached_plan}"

        treatment_plan = self.treatment_plans_fv.get_feature_vector(
            {"patient_id": patient_ssn}
        )
        if treatment_plan and len(treatment_plan) > 0:
            plan_content = treatment_plan[0]  # Extract the string from the list
            treatment_plan_cache.set(patient_ssn, plan_content)
            return f"Yes, the patient has an existing treatment plan:\n\n{plan_content}"
        else:
            return "No, the patient does not have an existing treatment plan."

    def handle_create_new_plan(
        self, monai_detection: str, patient_ssn: str, chat_history: List[Any]
    ) -> str:
        prompt_template = """
        Create a comprehensive treatment plan based on all available information about the patient.

        Patient SSN: {patient_ssn}

        MONAI Detection: {monai_detection}

        Patient History: {patient_history}

        Create a detailed treatment plan addressing all identified conditions and potential health risks. 
        ALWAYS Take into account the patient conditions and observations.
        Format the plan similar to this example:

        Patient X: **Patient ID: X - Treatment Plan**

        **Conditions:**
        1. Condition A
        2. Condition B

        **Treatment Goals:**
        - Goal 1
        - Goal 2

        **1. Condition A:**
           - **Treatment Step 1:** Details...
           - **Treatment Step 2:** Details...

        **2. Condition B:**
           - **Treatment Step 1:** Details...
           - **Treatment Step 2:** Details...

        **General Recommendations:**
        - Recommendation 1
        - Recommendation 2

        **Note:** Any additional notes or considerations.
        """

        patient_history = self.get_patient_history(patient_ssn, chat_history)
        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(
                monai_detection=monai_detection,
                patient_history=patient_history,
                patient_ssn=patient_ssn,
            )
        )

        new_plan = response.content.strip()
        treatment_plan_cache.set(patient_ssn, new_plan)
        return f"New treatment plan created:\n\n{new_plan}"

    def handle_update_existing_plan(
        self, monai_detection: str, patient_ssn: str, chat_history: List[Any]
    ) -> str:
        existing_plan = treatment_plan_cache.get(
            patient_ssn
        ) or self.treatment_plans_fv.get_feature_vector({"patient_id": patient_ssn})
        if not existing_plan:
            return "No existing treatment plan found. Please create a new plan."

        prompt_template = """
        Review the existing treatment plan and recent patient information to determine if updates are needed.
        ALWAYS Take into account the patient conditions and observations.

        Existing Treatment Plan:
        {existing_plan}

        MONAI Detection: {monai_detection}

        Recent Patient History: {recent_history}

        Have there been any changes in the patient's condition since their last primary care visit? If yes, describe these changes and create an updated treatment plan. If no changes, confirm that the current plan is still appropriate.
        """

        recent_history = self.get_patient_history(patient_ssn, chat_history)
        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(
                existing_plan=existing_plan,
                monai_detection=monai_detection,
                recent_history=recent_history,
            )
        )

        updated_plan = response.content.strip()
        treatment_plan_cache.set(patient_ssn, updated_plan)
        return updated_plan

    def handle_outline_plan_timeline(
        self, patient_ssn: str, chat_history: List[Any]
    ) -> str:
        existing_plan = treatment_plan_cache.get(
            patient_ssn
        ) or self.treatment_plans_fv.get_feature_vector({"patient_id": patient_ssn})
        if not existing_plan:
            return "No existing treatment plan found. Please create a new plan before outlining a timeline."

        prompt_template = """
        Based on the patient's current treatment plan and history, outline the treatment plan for the following time periods:
        a. Next 3 months
        b. Next 6 months
        c. Next 9 months
        d. Next 12 months

        Current Treatment Plan:
        {current_plan}

        Patient History: {patient_history}

        Provide a detailed timeline for each period, including any scheduled treatments, follow-ups, or expected milestones.
        ALWAYS Take into account the patient conditions and observations.
        """

        patient_history = self.get_patient_history(patient_ssn, chat_history)
        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(current_plan=existing_plan, patient_history=patient_history)
        )

        return response.content.strip()

    def handle_next_immediate_step(
        self, monai_detection: str, patient_ssn: str, chat_history: List[Any]
    ) -> str:
        existing_plan = treatment_plan_cache.get(
            patient_ssn
        ) or self.treatment_plans_fv.get_feature_vector({"patient_id": patient_ssn})
        if not existing_plan:
            existing_plan = "No existing treatment plan found."

        prompt_template = """
        Based on the patient's current treatment plan, recent history, and MONAI detection results, what is the next immediate step in the patient's care?

        Current Treatment Plan:
        {current_plan}

        MONAI Detection: {monai_detection}

        Recent Patient History: {recent_history}

        Provide a clear and concise description of the next immediate action or step in the patient's care plan, regardless of whether the plan is new or existing.
        ALWAYS Take into account the patient conditions and observations.
        """

        recent_history = self.get_patient_history(patient_ssn, chat_history)
        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(
                current_plan=existing_plan,
                monai_detection=monai_detection,
                recent_history=recent_history,
            )
        )

        return response.content.strip()

    def handle_example_treatment_plans(self, monai_detection: str) -> str:
        condition = self.extract_condition(monai_detection)
        if not condition:
            return "Could not determine the condition for example treatment plans. Please specify a valid condition."

        similar_patient_plans = self.get_similar_patient_plans(condition)

        prompt_template = """
        Provide example treatment plans for patients with {condition} based on the following similar cases:

        Similar Patient Plans:
        {similar_patient_plans}

        Summarize common treatment approaches and any notable variations in the treatment plans for patients with {condition}.
        """

        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(
                condition=condition,
                similar_patient_plans="\n\n".join(similar_patient_plans),
            )
        )

        return response.content.strip()

    def handle_compare_treatment_plans(
        self, patient_ssn: str, chat_history: List[Any]
    ) -> str:
        existing_plan = treatment_plan_cache.get(
            patient_ssn
        ) or self.treatment_plans_fv.get_feature_vector({"patient_id": patient_ssn})
        if not existing_plan:
            return "No existing treatment plan found. Please create a new plan before comparing."

        prompt_template = """
        Compare the patient's prior treatment plan with the expected treatment plan based on current information.

        Prior Treatment Plan:
        {prior_plan}

        Current Patient Information: {current_info}

        Provide a detailed comparison, highlighting any differences and explaining the rationale behind any changes in the treatment approach.
        """

        current_info = self.get_patient_history(patient_ssn, chat_history)
        prompt = PromptTemplate.from_template(prompt_template)
        response = self.llm(
            prompt.format(prior_plan=existing_plan, current_info=current_info)
        )

        return response.content.strip()

    def get_patient_history(self, patient_ssn: str, chat_history: List[Any]) -> str:
        return "\n".join(
            [msg.content for msg in chat_history if isinstance(msg, ToolMessage)]
        )

    @lru_cache(maxsize=100)
    def extract_condition(self, query: str) -> Optional[str]:
        conditions = [
            "atelectasis",
            "cardiomegaly",
            "consolidation",
            "edema",
            "enlarged_cardiomediastinum",
            "fracture",
            "lung_lesion",
            "lung_opacity",
            "no_finding",
            "pleural_effusion",
            "pleural_other",
            "pneumonia",
            "pneumothorax",
            "support_devices",
        ]

        prompt_template = ChatPromptTemplate.from_template("""
        Extract one medical condition from the following query. The condition should be one of: {conditions}

        Query: {query}

        If no specific condition is mentioned, respond with "unknown".

        Extracted condition:
        """)

        extraction_chain = LLMChain(llm=self.llm, prompt=prompt_template)
        extracted_condition = (
            extraction_chain.run(conditions=", ".join(conditions), query=query)
            .strip()
            .lower()
        )

        # Remove any "extracted condition:" prefix if present
        extracted_condition = extracted_condition.replace(
            "extracted condition:", ""
        ).strip()

        logger.info(f"🧪 Extracted condition: {extracted_condition}")
        return extracted_condition if extracted_condition in conditions else None

    def get_similar_patient_plans(self, condition: str) -> List[str]:
        print(f"🧬 Retrieved condition: {condition}")

        outcomes_df = self.outcomes_fv.get_batch_data()
        similar_patient_ids = outcomes_df[
            (outcomes_df[condition] == 1)
        ].patient_id.values[:5]

        print(f"👥 Similar patients: {similar_patient_ids}")

        # Hardcode patient IDs due to limited treatment plans in dataset (only 100 plans for patients 1-100)
        # This ensures consistent showcase results, as actual similar_patient_ids may exceed available plans
        similar_patient_ids = [10, 15, 17]

        similar_plans = []
        for pid in similar_patient_ids:
            plan = self.treatment_plans_fv.get_feature_vector({"patient_id": pid})
            if plan and len(plan) > 0:
                similar_plans.append(plan[0])

        return similar_plans


def clear_treatment_plan_cache():
    treatment_plan_cache.clear()
