import unittest
from unittest.mock import patch
import json
from functions.chatbot_api.agent import run_agent
import logging
from collections import defaultdict

# Set up logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


def log_separator(char="=", length=75):
    logger.info(char * length)


class TestMedicalAssistantSystem(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.assistant_stats = defaultdict(
            lambda: {"total": 0, "correct": 0, "wrong": 0, "no_tool": 0}
        )
        cls.wrong_invocations = []
        cls.no_tool_invoked = []

    def setUp(self):
        self.patcher1 = patch("app.st")
        self.mock_st = self.patcher1.start()

        self.mock_report = "Sample report text"
        self.mock_monai_detection = json.dumps({"detection": "Sample detection result"})
        self.mock_patient_ssn = "4"

    def check_tool_accuracy(self, result, expected_tool, question):
        stats = self.assistant_stats[expected_tool]
        stats["total"] += 1

        if (
            result is None
            or "output" not in result
            or isinstance(result["output"], str)
        ):
            logger.warning(f"❗ No tool was invoked for question: {question}")
            self.no_tool_invoked.append((question, expected_tool))
            stats["no_tool"] += 1
            return

        output = result["output"]
        used_tool = output.name if hasattr(output, "name") else "UnknownTool"

        if used_tool.lower() == expected_tool.lower():
            stats["correct"] += 1
        else:
            stats["wrong"] += 1
            self.wrong_invocations.append((question, expected_tool, used_tool))

    def run_test_for_questions(self, questions, expected_tool):
        for question in questions:
            with self.subTest(question=question):
                result = run_agent(
                    question,
                    self.mock_monai_detection,
                    self.mock_patient_ssn,
                    self.mock_report,
                )
                self.check_tool_accuracy(result, expected_tool, question)

    def test_assessment_questions(self):
        assessment_questions = [
            "Tell me everything I know about the patient",
            "What are the risk factors for this patient",
            "Based on the provided patient information, is there sufficient data to create a screening and workup plan?",
            "What limitations exist in the current patient information for creating a comprehensive screening and workup plan?",
            "What additional information would be most valuable to improve the screening and workup process?",
            "Should I still go ahead and do the screening if I cannot get the additional information?",
            "Using only the currently available information, what would be the best possible screening and workup plan?",
            "For the patient, are there any new observations or notes since their last visit",
            "What other questions should I ask the patient in order to make a more thorough assessment?",
            "Which tests should I order to check for a condition?",
        ]

        self.run_test_for_questions(assessment_questions, "AssessmentAssistant")

    def test_diagnosis_questions(self):
        diagnosis_questions = [
            "Do I have any observations for the patient",
            "Do I have enough information to diagnose any conditions? If so, what is the diagnosis? If not, what information do I need?",
            "What other questions should I ask the patient in order to make a more accurate diagnosis?",
        ]

        self.run_test_for_questions(diagnosis_questions, "DiagnosisAssistant")

    def test_prognosis_questions(self):
        prognosis_questions = [
            "Do I have enough information to make a reliable prognosis? If so, what is it? If not, what information do I need?",
            "Create a prognosis for my patient",
            "Compare the prognosis of a patient with the prognosis of similar patients with a particular attribute",
            "What other questions should I ask the patient in order to make a more accurate prognosis?",
        ]

        self.run_test_for_questions(prognosis_questions, "PrognosisAssistant")

    def test_treatment_plan_questions(self):
        treatment_plan_questions = [
            "Does the patient have a treatment plan?",
            "If no existing plan, create a comprehensive treatment plan, based on all available information about the patient",
            "If there is an existing treatment plan, have there been any changes in the patient's condition since their last primary care visit? If yes, describe these changes and create an updated treatment plan. If no changes, confirm the current plan is still appropriate.",
            "Outline the treatment plan for the following time periods: a. Next 3 months; b. Next 6 months; c. Next 9 months; d. Next 12 months",
            "What is the next immediate step in the patient's care, regardless of whether the plan is new or existing?",
            "What are some example treatment plans for patients where a particular condition is detected",
            "Compare the patient's prior treatment plan with expected treatment plan",
        ]

        self.run_test_for_questions(treatment_plan_questions, "TreatmentAssistant")

    def test_behavioral_insights_questions(self):
        behavioral_insights_questions = [
            "Can you show me the journey of a patient who has been diagnosed with pneumonia?",
            "I'd like to see the care pathway for a patient who has been readmitted to the hospital. Can you show me that?",
            "What are the most common reasons patients get readmitted to the hospital within 30 days?",
            "Can you show me patients with condition code 444814009?",
            "Compare similar patients with and without a pneumonia",
            "Compare patients with the lung opacity where the clinical notes are anomalous",
        ]

        self.run_test_for_questions(
            behavioral_insights_questions, "BehavioralQueryAssistant"
        )

    def tearDown(self):
        self.patcher1.stop()

    @classmethod
    def tearDownClass(cls):
        cls.print_system_summary()

    @classmethod
    def print_system_summary(cls):
        log_separator("=")
        logger.info("\n📊 System-wide Performance Summary")
        log_separator("-")

        total_correct = 0
        total_tests = 0

        logger.info(
            f"{'Assistant Type':<25} {'Accuracy':<10} {'Correct':<10} {'Wrong':<10} {'No Tool':<10} {'Total':<10}"
        )
        log_separator("-")

        for assistant, stats in cls.assistant_stats.items():
            accuracy = (
                (stats["correct"] / stats["total"]) * 100 if stats["total"] > 0 else 0
            )
            logger.info(
                f"{assistant:<25} {accuracy:>6.2f}%    {stats['correct']:<10} {stats['wrong']:<10} {stats['no_tool']:<10} {stats['total']:<10}"
            )
            total_correct += stats["correct"]
            total_tests += stats["total"]

        log_separator("-")
        overall_accuracy = (total_correct / total_tests) * 100 if total_tests > 0 else 0
        total_no_tool = sum(stats["no_tool"] for stats in cls.assistant_stats.values())
        logger.info(
            f"{'Overall':<25} {overall_accuracy:>6.2f}%    {total_correct:<10} {total_tests - total_correct - total_no_tool:<10} {total_no_tool:<10} {total_tests:<10}"
        )

        log_separator("=")
        logger.info("\n❌ Wrong Invocations:")
        log_separator("-")
        for question, expected_tool, used_tool in cls.wrong_invocations:
            logger.info(f"📝 Question: {question}")
            logger.info(f"🙏🏻 Expected Tool: {expected_tool}")
            logger.info(f"⛳️ Used Tool: {used_tool}")
            log_separator("-")

        if cls.no_tool_invoked:
            log_separator("=")
            logger.info("\n⚠️  Questions with No Tool Invoked:")
            log_separator("-")
            for question, expected_tool in cls.no_tool_invoked:
                logger.info(f"📝 Question: {question}")
                logger.info(f"🙏🏻 Expected Tool: {expected_tool}")
                log_separator("-")


if __name__ == "__main__":
    unittest.main()
